/* Required for strdup, snprintf, getopt */
#define _XOPEN_SOURCE 500
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <mntent.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/select.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

typedef struct sProperty
{
	char *name;
	char *value;
} Property;

typedef struct sDevice
{
	char *node;
	char *devname;
	char *label;
	char *description;
	char *mount_point;
	time_t time;
} Device;

int verbosity = 0;
char *post_mount_command = NULL;

/**
Parses a string of the form name=value and places the components in a Property
structure.  Returns 0 on success, or -1 if the string wasn't a valid property.
*/
int parse_property(char *str, int size, Property *prop)
{
	int equals = -1;
	int i;

	for(i=0; (equals<0 && i<size); ++i)
		if(str[i]=='=')
			equals = i;

	if(equals<0)
		return -1;

	prop->name = malloc(equals+1);
	strncpy(prop->name, str, equals);
	prop->name[equals] = 0;

	prop->value = malloc(size-equals);
	strncpy(prop->value, str+equals+1, size-equals-1);
	prop->value[size-equals-1] = 0;

	return 0;
}

/**
Retrieves all properties associated with a /dev node.  The returned array is
terminated with an entry containing NULL values.  Use free_properties to free
the array.
*/
Property *get_device_properties(char *node)
{
	int pid;
	int pipe_fd[2];
	char *buf;
	int bufsize;
	int pos = 0;
	int eof = 0;
	Property *props = NULL;
	int n_props = 0;

	pipe(pipe_fd);

	pid = fork();
	if(pid==0)
	{
		/* Child process */
		if(verbosity>=2)
			printf("Running udevadm info -q property -n \"%s\"\n", node);

		close(pipe_fd[0]);
		dup2(pipe_fd[1], 1);

		execl("/sbin/udevadm", "udevadm", "info", "-q", "property", "-n", node, NULL);
		_exit(1);
	}
	else if(pid<0)
	{
		close(pipe_fd[0]);
		close(pipe_fd[1]);

		return NULL;
	}

	/* Parent process */
	close(pipe_fd[1]);

	bufsize = 256;
	buf = (char *)malloc(bufsize);

	while(1)
	{
		int newline;
		int i;
		Property prop;

		if(!eof)
		{
			int len;

			len = read(pipe_fd[0], buf+pos, bufsize-pos);
			if(len==0)
				eof = 1;
			else if(len==-1)
				break;
			pos += len;
		}

		newline = -1;
		for(i=0; (newline<0 && i<pos); ++i)
			if(buf[i]=='\n')
				newline = i;

		if(newline<0)
		{
			if(eof)
				break;

			/* There was no newline in the buffer but there is more output to
			be read.  Try again with a larger buffer. */
			bufsize *= 2;
			buf = (char *)realloc(buf, bufsize);
			continue;
		}

		if(parse_property(buf, newline, &prop)==0)
		{
			/* Reserve space for a sentinel value as well. */
			props = (Property *)realloc(props, (n_props+2)*sizeof(Property));
			props[n_props] = prop;
			++n_props;

			memmove(buf, buf+newline+1, pos-newline-1);
			pos -= newline+1;
		}
		else
			break;
	}

	free(buf);

	if(props)
	{
		/* Terminate the array with NULL pointers. */
		props[n_props].name = NULL;
		props[n_props].value = NULL;
	}

	waitpid(pid, NULL, 0);
	close(pipe_fd[0]);

	return props;
}

/**
Looks for a property in an array of properties and returns its value.  Returns
NULL if the property was not found.
*/
char *get_property_value(Property *props, char *name)
{
	int i;
	for(i=0; props[i].name; ++i)
		if(strcmp(props[i].name, name)==0)
			return props[i].value;
	return NULL;
}

/**
Checks if a property has a specific value.  A NULL value is matched if the
property does not exist.
*/
int match_property_value(Property *props, char *name, char *value)
{
	char *v = get_property_value(props, name);
	if(!v)
		return value==NULL;
	return strcmp(v, value)==0;
}

/**
Frees an array of properties and all strings contained in it.
*/
void free_properties(Property *props)
{
	int i;
	if(!props)
		return;
	for(i=0; props[i].name; ++i)
	{
		free(props[i].name);
		free(props[i].value);
	}
	free(props);
}

/**
Returns an array of user-mountable devices listed in fstab.
*/
char **get_fstab_devices(void)
{
	FILE *file;
	struct mntent *me;
	char **devices = NULL;
	int n_devices = 0;

	file = setmntent("/etc/fstab", "r");
	if(!file)
		return NULL;

	while((me = getmntent(file)))
		if(hasmntopt(me, "user")!=NULL)
		{
			devices = (char **)realloc(devices, (n_devices+2)*sizeof(char *));
			devices[n_devices] = strdup(me->mnt_fsname);
			++n_devices;
		}

	endmntent(file);
	if(devices)
		devices[n_devices] = NULL;

	return devices;
}

/**
Checks if an array of strings contains the specified string.
*/
int is_in_array(char **array, char *str)
{
	int i;
	if(!array || !str)
		return 0;
	for(i=0; array[i]; ++i)
		if(!strcmp(str, array[i]))
			return 1;
	return 0;
}

/**
Frees an array of strings.
*/
void free_string_array(char **array)
{
	int i;
	if(!array)
		return;
	for(i=0; array[i]; ++i)
		free(array[i]);
	free(array);
}

/**
Checks if a partition identified by a sysfs path is on a removable device.
*/
int is_removable(char *devpath)
{
	char fnbuf[256];
	int len;
	char *ptr;
	int fd;

	len = snprintf(fnbuf, sizeof(fnbuf), "/sys%s", devpath);
	/* Default to not removable if the path was too long. */
	if(len+10>=(int)sizeof(fnbuf))
		return 0;

	/* We got a partition as a parameter, but the removable property is on the
	disk.  Replace the last component with "removable". */
	for(ptr=fnbuf+len; (ptr>fnbuf && *ptr!='/'); --ptr) ;
	strcpy(ptr, "/removable");

	fd = open(fnbuf, O_RDONLY);
	if(fd!=-1)
	{
		char c;
		read(fd, &c, 1);
		close(fd);
		if(c=='1')
		{
			if(verbosity>=2)
				printf("  Removable\n");
			return 1;
		}
		if(verbosity>=2)
			printf("  Not removable\n");
	}

	return 0;
}

/**
Checks if a partition's disk or any of its parent devices are connected to any
of a set of buses.  The device is identified by a sysfs path.  The bus array
must be terminated with a NULL entry.
*/
int check_buses(char *devpath, char **buses)
{
	char fnbuf[256];
	char *ptr;
	int len;

	len = snprintf(fnbuf, sizeof(fnbuf), "/sys%s", devpath);
	/* Default to no match if the path was too long. */
	if(len+10>=(int)sizeof(fnbuf))
		return 0;

	for(ptr=fnbuf+len; ptr>fnbuf+12; --ptr)
		if(*ptr=='/')
		{
			char linkbuf[256];
			/* Replace the last component with "subsystem". */
			strcpy(ptr, "/subsystem");
			len = readlink(fnbuf, linkbuf, sizeof(linkbuf)-1);

			if(len!=-1)
			{
				linkbuf[len] = 0;
				/* Extract the last component of the subsystem symlink. */
				for(; (len>0 && linkbuf[len-1]!='/'); --len) ;

				if(verbosity>=2)
				{
					*ptr = 0;
					printf("  Subsystem of %s is %s\n", fnbuf, linkbuf+len);
				}

				if(is_in_array(buses, linkbuf+len))
					return 1;
			}
		}

	return 0;
}

/**
Check if an array of properties describes a device that can be mounted.  An
array of explicitly allowed devices can be passed in as well.  Both arrays must
be terminated by a NULL entry.
*/
int can_mount(Property *props, char **allowed)
{
	static char *removable_buses[] = { "usb", "firewire", 0 };
	char *devname;
	char *devpath;
	char *bus;

	devname = get_property_value(props, "DEVNAME");
	if(is_in_array(allowed, devname))
		return 1;

	/* Special case for CD devices, since they are not partitions.  Only allow
	mounting if media is inserted. */
	if(match_property_value(props, "ID_TYPE", "cd") && match_property_value(props, "ID_CDROM_MEDIA", "1"))
		return 1;

	/* Only allow mounting partitions. */
	if(!match_property_value(props, "DEVTYPE", "partition"))
		return 0;

	devpath = get_property_value(props, "DEVPATH");
	if(is_removable(devpath))
		return 1;

	/* Certain buses are removable by nature, but devices only advertise
	themselves as removable if they support removable media, e.g. memory card
	readers. */
	bus = get_property_value(props, "ID_BUS");
	if(is_in_array(removable_buses, bus))
		return 1;

	return check_buses(devpath, removable_buses);
}

/**
Returns an array of all device nodes in a directory.  Symbolic links are
dereferenced.
*/
char **get_device_nodes(char *dirname)
{
	DIR *dir;
	struct dirent *de;
	char fnbuf[256];
	char linkbuf[256];
	struct stat st;
	char **nodes = NULL;
	int n_nodes = 0;
	char **checked = NULL;
	int n_checked = 0;
	int i;

	dir = opendir(dirname);
	if(!dir)
		return NULL;

	while((de = readdir(dir)))
	{
		char *node;
		int duplicate = 0;

		/* Ignore . and .. entries. */
		if(de->d_name[0]=='.' && (de->d_name[1]==0 || (de->d_name[1]=='.' && de->d_name[2]==0)))
			continue;

		snprintf(fnbuf, sizeof(fnbuf), "%s/%s", dirname, de->d_name);

		node = fnbuf;
		lstat(fnbuf, &st);
		if(S_ISLNK(st.st_mode))
		{
			int len;
			len = readlink(fnbuf, linkbuf, sizeof(linkbuf)-1);
			if(len!=-1)
			{
				linkbuf[len] = 0;
				node = linkbuf;
			}
		}

		/* There may be multiple symlinks to the same device.  Only include each
		device once in the returned array. */
		if(checked)
		{
			for(i=0; (!duplicate && i<n_checked); ++i)
				if(strcmp(node, checked[i])==0)
					duplicate = 1;
		}
		if(duplicate)
		{
			if(verbosity>=2)
				printf("Device %s is a duplicate\n", fnbuf);
			continue;
		}

		checked = (char **)realloc(checked, (n_checked+1)*sizeof(char *));
		checked[n_checked] = strdup(node);
		++n_checked;

		nodes = (char **)realloc(nodes, (n_nodes+2)*sizeof(char *));
		nodes[n_nodes] = strdup(fnbuf);
		++n_nodes;
	}

	closedir(dir);
	if(checked)
	{
		for(i=0; i<n_checked; ++i)
			free(checked[i]);
		free(checked);
	}

	if(nodes)
		nodes[n_nodes] = NULL;

	return nodes;
}

/**
Reads the list of mounted devices from /etc/mtab and records the mount points.
*/
void check_mounts(Device *devices)
{
	FILE *file;
	struct mntent *me;

	file = setmntent("/etc/mtab", "r");
	if(!file)
		return;

	while((me = getmntent(file)))
	{
		int i;

		for(i=0; devices[i].node; ++i)
			if(!strcmp(devices[i].devname, me->mnt_fsname))
			{
				devices[i].mount_point = strdup(me->mnt_dir);

				if(verbosity>=1)
					printf("Device %s is mounted on %s\n", devices[i].node, devices[i].mount_point);
			}
	}

	endmntent(file);
}

/**
Returns an array of all mountable devices.
*/
Device *get_devices(void)
{
	char **nodes = NULL;
	Device *devices = NULL;
	int n_devices = 0;
	char **mounted = NULL;
	char **fstab = NULL;
	int i;

	nodes = get_device_nodes("/dev/disk/by-id");
	fstab = get_fstab_devices();

	for(i=0; nodes[i]; ++i)
	{
		Property *props;

		if(verbosity>=1)
			printf("Examining device %s\n", nodes[i]);

		props = get_device_properties(nodes[i]);
		if(!props)
		{
			if(verbosity>=2)
				printf("  No properties\n");
			continue;
		}

		if(verbosity>=2)
		{
			int j;
			for(j=0; props[j].name; ++j)
				printf("  %s = %s\n", props[j].name, props[j].value);
		}

		if(can_mount(props, fstab))
		{
			char *devname;
			char *label;
			char *vendor;
			char *model;
			char buf[256];
			char pos;
			struct stat st;

			if(verbosity>=1)
				printf("  Using device\n");

			devname = get_property_value(props, "DEVNAME");

			/* Get a human-readable label for the device.  Use filesystem label,
			filesystem UUID or device node name in order of preference. */
			label = get_property_value(props, "ID_FS_LABEL");
			if(!label)
				label = get_property_value(props, "ID_FS_UUID");
			if(!label)
			{
				char *ptr;

				label = devname;
				for(ptr=label; *ptr; ++ptr)
					if(*ptr=='/')
						label = ptr+1;
			}

			vendor = get_property_value(props, "ID_VENDOR");
			model = get_property_value(props, "ID_MODEL");

			pos = snprintf(buf, sizeof(buf), "%s", label);
			if(vendor && model)
				pos += snprintf(buf+pos, sizeof(buf)-pos, " (%s %s)", vendor, model);

			stat(nodes[i], &st);

			/* Reserve space for a sentinel entry. */
			devices = (Device *)realloc(devices, (n_devices+2)*sizeof(Device));
			devices[n_devices].node = nodes[i];
			devices[n_devices].devname = strdup(devname);
			devices[n_devices].label = strdup(label);
			devices[n_devices].description = strdup(buf);
			devices[n_devices].mount_point = NULL;
			devices[n_devices].time = st.st_mtime;
			++n_devices;
		}
		else
			free(nodes[i]);
		free_properties(props);
	}

	free(nodes);
	free_string_array(mounted);

	if(devices)
	{
		/* Terminate the array with NULL pointers. */
		devices[n_devices].node = NULL;
		devices[n_devices].devname = NULL;
		devices[n_devices].label = NULL;
		devices[n_devices].description = NULL;
		devices[n_devices].mount_point = NULL;

		check_mounts(devices);
	}

	return devices;
}

/**
Frees an array of devices and all strings contained in it.
*/
void free_devices(Device *devices)
{
	int i;
	if(!devices)
		return;
	for(i=0; devices[i].node; ++i)
	{
		free(devices[i].node);
		free(devices[i].devname);
		free(devices[i].label);
		free(devices[i].description);
		if(devices[i].mount_point)
			free(devices[i].mount_point);
	}
	free(devices);
}

/**
Mounts a device if it was not mounted, or unmounts if it was.
*/
int toggle_device(Device *device, char *out_buf, int out_size)
{
	int umount = !!device->mount_point;
	char mount_point[1024];
	int pos = 0;
	int status = 0;
	fd_set fds;
	struct timeval timeout;
	int pid;
	int pipe_fd[2];
	char suffix = 0;

	out_buf[0] = 0;

	/* Find a mount point that does not exist yet. */
	while(1)
	{
		int len;

		len = snprintf(mount_point, sizeof(mount_point), "/media/%s", device->label);
		if(len+2>=(int)sizeof(mount_point))
			return -1;

		if(suffix)
			len += snprintf(mount_point+len, sizeof(mount_point)-len, "_%c", suffix);

		if(access(mount_point, F_OK)<0 && errno==ENOENT)
			break;

		if(suffix==0)
			suffix = '1';
		else if(suffix<'9')
			++suffix;
		else
			return -1;
	}

	pipe(pipe_fd);

	pid = fork();
	if(pid==0)
	{
		/* Child process */
		if(verbosity>=1)
		{
			if(umount)
				printf("Running pumount %s\n", device->node);
			else
				printf("Running pmount %s %s\n", device->node, mount_point+7);
		}

		close(pipe_fd[0]);
		dup2(pipe_fd[1], 1);
		dup2(pipe_fd[1], 2);

		if(umount)
			execl("/usr/bin/pumount", "pumount", device->node, NULL);
		else
			execl("/usr/bin/pmount", "pmount", device->node, mount_point+7, NULL);
		_exit(1);
	}
	else if(pid<0)
		return -1;

	/* Parent process */

	close(pipe_fd[1]);
	FD_ZERO(&fds);
	FD_SET(pipe_fd[0], &fds);
	timeout.tv_sec = 0;
	timeout.tv_usec = 200000;

	while(1)
	{
		/* The write fd for the pipe may be inherited by a fuse server
		process and stay open indefinitely. */
		if(select(pipe_fd[0]+1, &fds, NULL, NULL, &timeout))
		{
			int len;

			len = read(pipe_fd[0], out_buf+pos, out_size-pos-1);
			if(len<=0)
				break;
			pos += len;
		}
		else if(waitpid(pid, &status, 0))
		{
			pid = 0;
			break;
		}
	}

	close(pipe_fd[0]);
	if(pid)
		waitpid(pid, &status, 0);

	out_buf[pos] = 0;

	if(verbosity>=1)
	{
		if(WIFEXITED(status))
		{
			if(WEXITSTATUS(status))
				printf("Command exited successfully\n");
			else
				printf("Command exited with status %d\n", WEXITSTATUS(status));
		}
		else if(WIFSIGNALED(status))
			printf("Command terminated with signal %d\n", WTERMSIG(status));
		else
			printf("Command exited with unknown result %04X\n", status);
	}

	if(!WIFEXITED(status) || WEXITSTATUS(status))
		return -1;

	if(umount)
	{
		free(device->mount_point);
		device->mount_point = NULL;
	}
	else
		device->mount_point = strdup(mount_point);

	return 0;
}

/**
Callback for activating a row in the device list.  Mounts or unmounts the
device depending on operating mode.
*/
void row_activated(GtkTreeView *list, GtkTreePath *path, GtkTreeViewColumn *column, gpointer user_data)
{
	GtkTreeModel *model;
	GtkTreeIter iter;
	Device *device;
	int ret;
	char output[1024];
	int pid;

	model = gtk_tree_view_get_model(list);

	if(!gtk_tree_model_get_iter(model, &iter, path))
		return;

	gtk_tree_model_get(model, &iter, 1, &device, -1);
	ret = toggle_device(device, output, sizeof(output));
	if(ret)
	{
		GtkWidget *dialog;

		/* Pmount terminated with nonzero status or a signal.  Display an
		error to the user. */
		dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "%s", output);
		g_signal_connect(dialog, "response", &gtk_main_quit, NULL);
		gtk_widget_show_all(dialog);
		return;
	}

	gtk_main_quit();

	if(post_mount_command && device->mount_point)
	{
		if(verbosity>=1)
			printf("Running %s in %s\n", post_mount_command, device->mount_point);

		pid = fork();
		if(pid==0)
		{
			chdir(device->mount_point);
			execlp(post_mount_command, post_mount_command, NULL);
			_exit(1);
		}
	}

	(void)column;
	(void)user_data;
}

/**
Callback for the mount/unmount button.  Causes the selected row in the device
list to be activated.
*/
void button_clicked(GtkButton *button, gpointer user_data)
{
	GtkWidget *list = (GtkWidget *)user_data;
	GtkTreeSelection *selection;
	GtkTreeIter iter;
	GtkTreeModel *model;
	GtkTreePath *path;

	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(list));
	gtk_tree_selection_get_selected(selection, &model, &iter);
	path = gtk_tree_model_get_path(model, &iter);
	gtk_tree_view_row_activated(GTK_TREE_VIEW(list), path, gtk_tree_view_get_column(GTK_TREE_VIEW(list), 0));
	gtk_tree_path_free(path);

	(void)button;
}

/**
Global key press callback for the window.
*/
gboolean key_press(GtkWidget *widget, GdkEvent *event, gpointer user_data)
{
	if(event->key.keyval==GDK_KEY_Escape)
	{
		gtk_main_quit();
		return TRUE;
	}

	(void)widget;
	(void)user_data;

	return FALSE;
}

void show_help(void)
{
	printf("pmount-gui\n"
		"Copyright (c) 2011-2015 Mikko Rasa, Mikkosoft Productions\n\n"
		"Usage: pmount-gui [-v] [-u] [-r <command>] [-h]\n\n"
		"Options:\n"
		"  -v  Increase verbosity\n"
		"  -u  Unmount a device (default is mount)\n"
		"  -r  Run a command after mounting\n"
		"  -h  Display this help\n");
}

int main(int argc, char **argv)
{
	GtkWidget *window;
	GtkWidget *box;
	GtkWidget *viewport;
	GtkWidget *list;
	GtkListStore *store;
	GtkTreeSelection *selection;
	GtkWidget *button;
	GtkTreeIter iter;
	Device *devices;
	int i;
	time_t latest;
	int opt;
	int umount = 0;
	int n_listed;

	gtk_init(&argc, &argv);

	while((opt = getopt(argc, argv, "vur:h"))!=-1) switch(opt)
	{
	case 'v':
		++verbosity;
		break;
	case 'u':
		umount = 1;
		break;
	case 'r':
		post_mount_command = optarg;
		break;
	case 'h':
		show_help();
		return 0;
	}

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_container_set_border_width(GTK_CONTAINER(window), 5);
	g_signal_connect(window, "destroy", G_CALLBACK(&gtk_main_quit), NULL);
	g_signal_connect(window, "key-press-event", G_CALLBACK(&key_press), NULL);

	box = gtk_vbox_new(FALSE, 5);
	gtk_container_add(GTK_CONTAINER(window), box);

	viewport = gtk_viewport_new(NULL, NULL);
	gtk_viewport_set_shadow_type(GTK_VIEWPORT(viewport), GTK_SHADOW_IN);
	gtk_box_pack_start(GTK_BOX(box), viewport, TRUE, TRUE, 0);

	list = gtk_tree_view_new();
	gtk_container_add(GTK_CONTAINER(viewport), list);
	g_signal_connect(list, "row-activated", G_CALLBACK(&row_activated), NULL);

	store = gtk_list_store_new(2, G_TYPE_STRING, G_TYPE_POINTER);
	gtk_tree_view_set_model(GTK_TREE_VIEW(list), GTK_TREE_MODEL(store));
	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(list),
		-1, "Device", gtk_cell_renderer_text_new(), "text", 0, NULL);

	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(list));

	button = gtk_button_new_with_label(umount ? "Unmount" : "Mount");
	g_signal_connect(button, "clicked", G_CALLBACK(&button_clicked), list);
	gtk_box_pack_start(GTK_BOX(box), button, FALSE, TRUE, 0);

	devices = get_devices();
	n_listed = 0;
	if(devices)
	{
		/* Populate the list with devices in appropriate state. */
		latest = 0;
		for(i=0; devices[i].node; ++i)
			if(!devices[i].mount_point==!umount)
			{
				gtk_list_store_append(store, &iter);
				gtk_list_store_set(store, &iter, 0, devices[i].description, 1, &devices[i], -1);
				if(devices[i].time>latest)
				{
					/* Pre-select the device that appeared on the system most recently. */
					latest = devices[i].time;
					gtk_tree_selection_select_iter(selection, &iter);
				}

				++n_listed;
			}

	}

	if(n_listed)
		gtk_widget_show_all(window);
	else
	{
		GtkWidget *dialog;

		/* Don't show the main window if no devices were listed. */
		dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_WARNING, GTK_BUTTONS_OK,
			"No devices to %s", (umount ? "unmount" : "mount"));
		g_signal_connect(dialog, "response", G_CALLBACK(&gtk_main_quit), NULL);
		gtk_widget_show_all(dialog);
	}

	gtk_main();

	free_devices(devices);

	return 0;
}
